
/******************************************************************************
* MODULE     : evaluate.cc
* DESCRIPTION: The main evaluation routines
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <single.hh>
#include <literal.hh>
#include <xtable.hh>
#include <keyword.hh>
#include <error.hh>
#include <function.hh>
namespace mmx {

/******************************************************************************
* The main loop
******************************************************************************/

generic
evaluator_rep::loop () {
  while (!nil (cur)) {
    //cout << "Trace " << trace << lf;
    //if (!nil (trace)) cout << "Evaluate " << car(trace) << ", " << cur << lf;
    //cout << "Evaluate " << N(trace) << ", " << cur << lf;
    //cout << "Evaluate " << cur << lf;
    //cout << "Evaluate " << cur << ", " << rem << lf;
    //cout << "Environment " << env->local_level() << ", " <<N(env) << ", " << ((int) env.rep) << lf;
    switch (gentype (cur)) {
    case STRING_TYPE: {
      generic tmp= cur;
      cur= car (rem);
      rem= cons (value (tmp, ID_STRING), cdr (rem));
      break;
    }
    case SYMBOL_TYPE:
      if (env->contains (cur)) {
	generic tmp= cur;
	cur= car (rem);
	rem= cons (env [tmp], cdr (rem));
	//cout << "  --> " << car (rem) << lf;
      }
      else {
	// cout << "env= " << env << lf;
	error (cur, "Unknown identifier");
      }
      break;
    case LIST_TYPE:
      trace= cons (cur, trace);
      rem  = cons (KW_COMPOUND_1, rem);
      cur  = car (cur);
      break;
    case SINGLE_TYPE: {
      generic tmp= cur;
      cur= car (rem);
      rem= cons (value (tmp, ID_INT), cdr (rem));
      break;
    }
    case VALUE_TYPE: {
      generic tmp= cur;
      cur= car (rem);
      rem= cons (tmp, cdr (rem));
      break;
    }
    case LITERAL_INT_TYPE: {
      generic g= env [MMX_NEW_INT];
      generic v= value (cur, ID_LITERAL_INT);
      if (nil (g))
	error (cur, "No method for literal integers available");
      else {
	trace= cons (cur, trace);
	cur= KW_COMPOUND_2;
	rem= cons (v, list (g), list (), rem);
	kw_compound_2 ();
      }
      break;
    }
    case LITERAL_FLOAT_TYPE: {
      generic g= env [MMX_NEW_FLOAT];
      generic v= value (cur, ID_LITERAL_FLOAT);
      if (nil (g))
	error (cur, "No method for literal floats available");
      else {
	trace= cons (cur, trace);
	cur= KW_COMPOUND_2;
	rem= cons (v, list (g), list (), rem);
	kw_compound_2 ();
      }
      break;
    }
    case LITERAL_STRING_TYPE: {
      generic g= env [MMX_NEW_STRING];
      generic v= value (cur, ID_LITERAL_STRING);
      if (nil (g))
	error (cur, "No method for literal strings available");
      else {
	trace= cons (cur, trace);
	cur= KW_COMPOUND_2;
	rem= cons (v, list (g), list (), rem);
	kw_compound_2 ();
      }
      break;
    }
    case KEYWORD_TYPE:
      (this ->* (as_keyword (cur)->fn)) ();
      break;
    default:
      error (cur, "Syntax error");
    }
  }
  //cout << "env= " << env << lf;
  return car (rem);
}

/******************************************************************************
* Evaluating expressions
******************************************************************************/

generic
evaluator_rep::evaluate (generic g) {
  env (MISC_TODO)= list ();
  env (MISC_DONE)= list ();
  trace= list ();
  cur= g;
  rem= list (generic ());
  generic ret= loop ();
#ifdef DEBUG_ON
  while (env->local_level () != 0) {
    xtable t= as_xtable (env);
    cout << "Pending local variables" << lf << hrule;
    dump (t->ext_t);
    cout << hrule;
    env= t->base_t;
  }
#endif
  return ret;
}

void
evaluator_rep::error (generic where, string message) {
#ifdef DEBUG_ON
  cout << "cont = " << cons (cur, rem) << lf;
  cout << "trace= " << trace << lf;
#endif
  generic err= std_error (message, where);
  cur= generic ();
  rem= cons (list (symbol ("errors!"), err), rem);
}

/******************************************************************************
* The user interface
******************************************************************************/

evaluator mmeval;

generic
mmx_evaluate (generic g) {
  return mmeval->evaluate (g);
}

void
mmx_display (value v) {
  return mmeval->display (v);
}

void
extern_declare (symbol var_name, generic val, generic type) {
  mmeval->declare (var_name, value (val, type));
}

void
extern_declare (symbol var_name, mmx_function fn, generic type) {
  mmeval->declare (var_name, value (function (var_name, fn), type));
}

} // namespace mmx
