
/******************************************************************************
* MODULE     : ext_glue.cc
* DESCRIPTION: Generate glue with Mathemagix
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <single.hh>
namespace mmx {

/******************************************************************************
* Frequently used names and subroutines
******************************************************************************/


static generic
scope_mmx (generic g) {
  return Scope (symbol ("mmx"), g);
}

static generic mmx_generic= scope_mmx (symbol ("generic"));
static generic mmx_mirror = scope_mmx (symbol ("mirror"));
static generic mmx_symbol = scope_mmx (symbol ("symbol"));

static symbol
ext_id (int i) {
  return string ("ID_") * as_string (i);
}

static symbol
ext_fun (int i) {
  return string ("FUN_") * as_string (i);
}

static int nr_ext_tvar= 0;

static symbol
ext_tvar (int i) {
  if (i > nr_ext_tvar) nr_ext_tvar= i;
  return string ("TMPL_") * as_string (i);
}

static symbol
ext_arg (int i) {
  return string ("x_") * as_string (i);
}

void
extender::declare (generic var, generic val, generic type) {
  generic dc  = scope_mmx (symbol ("extern_declare"));
  generic decl= list (dc, var, val, type);
  glue_init= cons (decl, glue_init);
}

/******************************************************************************
* Building Mathemagix types in glue initialization
******************************************************************************/

static int
find_in_bindings (generic g, list l, int i=1) {
  if (nil (l)) return -1;
  if (is_func (car (l), MMX_TYPE, 2) && (car (l) [1] == g)) return i;
  return find_in_bindings (g, cdr (l), i+1);
}

bool
extender::match_mmx_type (generic expr, generic defn, list bindings) {
  // cerr << expr << " == " << defn << lf;
  if (expr == defn) return true;
  int i= find_in_bindings (expr, bindings);
  if (i != -1) return match_mmx_type (bindings[i-1][2], defn, list ());
  if ((!is_list (expr)) || nil (expr)) return false;
  if ((!is_list (defn)) || nil (defn)) return false;
  return
    match_mmx_type (car (expr), car (defn), bindings) &&
    match_mmx_type (cdr (expr), cdr (defn), bindings);
}

int
extender::resolve_mmx_type (generic expr, list l, int i, list bindings) {
  // FIXME: this should really be done using the converter from logic
  // FIXME: moreover the searching is only quadratic in time
  if (nil (l)) return -1;
  generic defn= car (l);
  if (is_func (defn, MMX_DEFINE, 2)) defn= defn[1];
  if (is_func (defn, MMX_TYPE, 2)) {
    if (is_func (defn[2], MMX_FUNCTION, 2)) {
      if (is_func (defn[2][1], MMX_CROSS))
	defn= cons (defn[1], cdr (as_list (defn[2][1])));
    }
    else defn= defn[1];
  }
  if (match_mmx_type (expr, defn, bindings)) return i;
  return resolve_mmx_type (expr, cdr (l), i+1, bindings);
}

generic
extender::build_mmx_type_sub (generic g, list bindings) {
  if (g == MMX_CLASS) return scope_mmx (symbol ("ID_CLASS"));
  if (g == MMX_VOID) return scope_mmx (symbol ("ID_VOID"));
  if (g == MMX_BOOL) return scope_mmx (symbol ("ID_BOOL"));
  if (g == MMX_ALIAS) return scope_mmx (symbol ("ID_ALIAS"));
  if (g == MMX_CROSS) return scope_mmx (symbol ("ID_CROSS"));
  if (g == MMX_FUNCTION) return scope_mmx (symbol ("ID_FUNCTION"));
  int i= find_in_bindings (g, bindings);
  if (i != -1) return ext_tvar (i);
  return g;
}

list
extender::build_mmx_type (list l, list bindings) {
  if (nil (l)) return l;
  return cons (build_mmx_type (car (l), bindings),
	       build_mmx_type (cdr (l), bindings));
}

generic
extender::build_mmx_type (generic g, list bindings) {
  if (N(g) == 0) {
    generic id= build_mmx_type_sub (g, bindings);
    if (id != g) return id;
    int i= resolve_mmx_type (g, types, 1, bindings);
    if (i != -1) return ext_id (i);
  }
  else {
    generic id= build_mmx_type_sub (g[0], bindings);
    if (id == g[0]) {
      int i= resolve_mmx_type (g, types, 1, bindings);
      if (i != -1) id= ext_id (i);
    }
    if (id != g[0]) {
      generic mmxl= scope_mmx (symbol ("list"));
      generic l   = build_mmx_type (cdr (as_list (g)), bindings);
      return cons (mmxl, id, l);
    }
  }
  error (g, "Expected a Mathemagix type");
  // cerr << "Bindings were " << bindings << lf;
  return g;
}

/******************************************************************************
* Produce the type constructing routines
******************************************************************************/

void
extender::produce_glue_type (generic g, int i) {
  // cerr << "Glue type " << g << "\n";
  if (!is_func (g, MMX_DEFINE, 2))
    error (g, "Expected a declaration");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1][1]))
    error (g[1][1], "Expected a symbol");
  else if (!is_symbol (g[2]))
    error (g[2], "Expected a symbol");
  else {
    generic id= ext_id (i);
    generic vv= list (mmx_mirror, list (mmx_symbol, S(g[1][1])));
    generate (Assign (Type (id, Static (mmx_generic)), vv));
    if (is_func (g[1][2], MMX_FUNCTION, 2)) {
      generic x  = symbol ("g");
      generic fun= ext_fun (i);
      generic arg= Type (x, mmx_generic);
      generic tp = Static (Function (arg, mmx_generic));
      generic l  = list (symbol ("cdr"), list (symbol ("as_list"), x));
      generic ret= list (symbol ("cons"), id, l);
      generate (Define (Type (fun, tp), Begin (Return (ret))));
      declare (S(g[1][1]), fun, build_mmx_type (g[1][2], list ()));
    }
    else declare (S(g[1][1]), id, build_mmx_type (g[1][2], list ()));
  }
}

void
extender::produce_glue_type (list l, int i) {
  if (!nil (l)) {
    produce_glue_type (car (l), i);
    produce_glue_type (cdr (l), i+1);
  }
}

/******************************************************************************
* Deal with template arguments
******************************************************************************/

generic
extender::view_as_ccpp (generic expr, generic mmx_type, table t) {
  generic type= type_as_ccpp (mmx_type, t);
  generic cv  = scope_mmx (Specialize (symbol ("view_as"), type));
  if (type == MMX_CLASS) {
    cv  = scope_mmx (Specialize (symbol ("view_as"), gentype));
    expr= list (scope_mmx (symbol ("as_glue_type")), expr);
  }
  return list (cv, expr);
}

list
extender::glue_template_args (list l, int kind, int i, int n) {
  if (nil (l)) return l;
  generic g= car (l);
  list res;
  if (is_func (g, MMX_TYPE, 2)) {
    generic var = g[1];
    generic type= g[2];
    generic tp  = type_as_ccpp (type);
    generic org = Access (symbol ("g"), single (i));
    symbol  arg = string ("x_") * as_string (i);
    symbol  save= string ("save_") * as_string (n);
    switch (kind) {
    case 1: // retrieve template arguments
      if (type == MMX_CLASS) {
	generic sv= Assign (Type (save, gentype), curtype (n));
	generic nv= Assign (curtype (n), view_as_ccpp (org, type));
	res= list (sv, nv);
      }
      else res= Assign (Type (arg, tp), view_as_ccpp (org, type));
      break;
    case 2: // get new names of template arguments
      if (type == MMX_CLASS) res= list (generic (list (var, geninst (n))));
      else res= list (generic (list (var, arg)));
      break;
    case 3: // restore generic types
      if (type == MMX_CLASS) res= list (Assign (curtype (n), save));
      break;
    }
    if (type == MMX_CLASS) n++;
  }
  return res * glue_template_args (cdr (l), kind, i+1, n);
}

/******************************************************************************
* Produce the main routines
******************************************************************************/

list
extender::glue_convert_args (list l, generic x, table t, int i) {
  if (nil (l)) return l;
  generic gi = Access (x, single (i));
  generic cgi= view_as_ccpp (gi, car (l), t);
  return cons (cgi, glue_convert_args (cdr (l), x, t, i+1));
}

void
extender::produce_glue_routine (generic g, int i, list bindings) {
  // cerr << "Glue routine " << g << "\n";
  if (is_func (g, MMX_FORALL, 2))
    produce_glue_routine (g[2], i, bindings * list (g[1]));
  else if (!is_func (g, MMX_DEFINE, 2))
    error (g, "Expected a declaration");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[2]))
    error (g[2], "Expected a symbol");
  else if (!is_func (g[1][2], MMX_FUNCTION, 2))
    error (g, "Constant exports not yet implemented");
  else if (!is_func (g[1][2][1], MMX_CROSS))
    error (g[1][2][1], "Cross expected");
  else {
    list code;
    generic x  = symbol ("g");
    generic fun= ext_fun (i);
    generic arg= Type (x, mmx_generic);
    generic tp = Static (Function (arg, mmx_generic));

    // Pretreatment of template arguments
    if (!nil (bindings)) {
      generic y = symbol ("h");
      generic ac= Assign (Type (y, mmx_generic), Access (x, single (0)));
      code= cons (ac, glue_template_args (bindings, 1));
      x= y;
    }
    table t= table_from_entries (glue_template_args (bindings, 2));

    // Execute main function and posttreatment of template arguments
    list    orga= cdr (as_list (g[1][2][1]));
    list    args= as_list (glue_convert_args (orga, x, t, 1));
    generic op  = ccpp_back (g[2]);
    generic ret = cons (op, args);
    generic rtp = type_as_ccpp (g[1][2][2], t);
    generic rcv = scope_mmx (Specialize (symbol ("glued"), rtp));
    if (rtp == MMX_VOID) {
      code= code * list (ret);
      code= code * glue_template_args (bindings, 3);
      code= code * list (Return (list (mmx_generic)));
    }
    else {
      if (nil (bindings)) code= code * list (Return (list (rcv, ret)));
      else {
	generic r= symbol ("ret");
	code= code * list (Assign (Type (r, mmx_generic), list (rcv, ret)));
	code= code * glue_template_args (bindings, 3);
	code= code * list (Return (r));
      }
    }
    generate (Define (Type (fun, tp), Begin (code)));

    // Put templates around declaration
    int j, n= N(bindings);
    generic dtp= build_mmx_type (g[1][2], bindings);
    for (j=n; j>0; j--) {
      generic mmxl= scope_mmx (symbol ("list"));
      dtp= list (mmxl, scope_mmx (symbol ("ID_FORALL")), ext_tvar (j), dtp);
    }
    declare (S(g[1][1]), fun, dtp);
  }
}

void
extender::produce_glue_routine (list l, int i) {
  if (!nil (l)) {
    produce_glue_routine (car (l), i, list ());
    produce_glue_routine (cdr (l), i+1);
  }
}

/******************************************************************************
* Main routine for producing all glue
******************************************************************************/

void
extender::produce_glue () {
  int i;
  nr_ext_tvar= 0;
  produce_glue_type (types, 1);
  produce_glue_routine (exports, N(types) + 1);
  for (i=1; i<=nr_ext_tvar; i++) {
    generic var= ext_tvar (i);
    generic val= list (mmx_mirror, list (mmx_symbol, var));
    generate (Assign (Type (var, mmx_generic), val));
  }
  generic var= symbol ("init_" * the_prefix);
  generic tp = list (CCPP_EXTERN, string ("C"), Function (MMX_VOID, MMX_VOID));
  generic val= Begin (reverse (glue_init));
  generate (Define (Type (var, tp), val));
}

} // namespace mmx
