
/******************************************************************************
* MODULE     : extend.hh
* DESCRIPTION: Convert mathemagix trees to C or C++
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __EXTEND_HH
#define __EXTEND_HH
#include <symbols.hh>
#include <list.hh>
namespace mmx {

class extender {
private:
  int    nr_errors;
  string preamble;
  string postamble;
  list   types;
  table  type_conv;
  list   exports;
  table  export_env;
  list   imports;
  list   output;
  int    max_type_args;
  list   glue_init;

  string  the_prefix;
  symbol  gentype;
  symbol  gentype_rep;
  generic genmethod (generic name);
  generic geninst ();
  generic geninst (int nr);
  generic curtype (int nr);
  generic contype (generic name);
  generic contype_rep (generic name);

private:
  void    error (generic where, string message);
  void    warning (generic where, string message);
  void    generate (generic g);
  generic rewrite (generic g);
  void    type_decl (generic g);
  void    export_decl (generic g);

  generic specialize (generic g, generic to, bool virtual_flag);
  list    specialize (list l, generic to, bool virtual_flag);
  generic type_as_ccpp (generic g, table t= table ());
  list    type_as_ccpp (list l, table t= table ());
  generic decl_as_ccpp (generic g, bool gen_flag= false);
  list    decl_as_ccpp (list l, bool gen_flag= false);
  void    gentype_methods (generic g);
  void    gentype_methods (list l);
  void    produce_gentype ();
  void    produce_geninst ();
  list    abstract (int nr, list args, list types);
  void    geninst_methods (int nr, generic g);
  void    geninst_methods (int nr, list l);
  void    produce_geninst (int nr);

  list    filter (list l, list target_l);
  list    class_to_geninst (list l, int i=1);
  list    contype_tmpl_args (list l, int kind, int i=1, int n=1);
  list    contype_cast_args (generic tp, list l, int i);
  void    contype_routine (generic n, generic t, list a, generic g, generic h);
  void    contype_produce (generic mmx_id, generic ccpp_id, list targs);
  void    produce_contype (generic g);
  void    produce_contype (list l);

  void    declare (generic var, generic val, generic type);
  bool    match_mmx_type (generic expr, generic defn, list bindings);
  int     resolve_mmx_type (generic expr, list l, int i, list bindings);
  generic build_mmx_type_sub (generic g, list bindings);
  list    build_mmx_type (list l, list bindings);
  generic build_mmx_type (generic g, list bindings);
  void    produce_glue_type (generic g, int i);
  void    produce_glue_type (list l, int i);
  generic view_as_ccpp (generic expr, generic mmx_type, table t= table ());
  list    glue_template_args (list l, int kind, int i=1, int n=1);
  list    glue_convert_args (list l, generic x, table t, int i);
  void    produce_glue_routine (generic g, int i, list bindings);
  void    produce_glue_routine (list l, int i);
  void    produce_glue ();

public:
  inline extender (string prefix):
    nr_errors (0),
    max_type_args (0),
    the_prefix (prefix),
    gentype (the_prefix * "_type"),
    gentype_rep (the_prefix * "_type_rep") {}
  inline ~extender () {}

  void parse (generic prog);
  void produce ();

  friend bool extension_code (ostream& out, string prefix, generic p, bool g);
};

inline symbol operator * (symbol s1, symbol s2) {
  return symbol (as_string (S (s1)) * as_string (S (s2))); }
void nice_print (generic text, list l);
void nice_print (symbol text, list l);

generic get_decl_var (generic g);
generic get_decl_type (generic g);
generic get_decl_val (generic g);
generic make_struct (generic name, list args, list inherit= list ());
generic make_blackbox_struct (generic name);
symbol  parameter (int i);
generic name_parameters (generic name);

}
#endif // __EXTEND_HH
