
/******************************************************************************
* MODULE     : glued.cc
* DESCRIPTION: Externally glued objects
* COPYRIGHT  : (C) 2004  Bernard Mourrain and Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_GLUED_HH
#define __MMX_GLUED_HH
#include <symbol.hh>
namespace mmx {

typedef generic (*mmx_function) (generic);
void extern_declare (symbol var_name, generic val, generic type);
void extern_declare (symbol var_name, mmx_function fn, generic type);
void extern_global (void *addr);
// defined in evaluate.cc

template<typename R> struct glued;
// template<typename R> ostream& operator << (ostream&, glued<R>);
template<typename R> glued<R> as_glued (generic g);

template<typename R>
struct glued_rep: generic_rep {
public:
  R data; // the glued data

protected:
  int Type ();
  // void Print (ostream& out);
  // int Hash ();
  // bool Equal (generic g) const;
  // generic Copy ();

public:
  inline glued_rep (R r): data (r) {}

  friend struct glued<R>;
  // friend ostream& operator << <> (ostream& out, glued<R> a);
  friend glued<R> as_glued<R> (generic g);
};

template<typename R>
struct glued {
protected:
  glued_rep<R>* rep;
  inline glued<R> (glued_rep<R>* r): rep (r) {}

public:
  inline glued<R> (R r): rep (new glued_rep<R> (r)) {}
  inline operator R () { return rep->rep; }
  inline operator generic () { return generic (rep); }
  inline glued_rep<R>* operator -> () { return rep; }
  inline bool operator == (glued<R> g) { return (rep->data == g->data); }
  inline bool operator != (glued<R> g) { return (rep->data != g->data); }

  friend struct glued_rep<R>;
  // friend ostream& operator << <> (ostream& out, glued<R> x);
  friend glued<R> as_glued<R> (generic g);
};

/******************************************************************************
* Required routines for derived classes from generic_rep
******************************************************************************/

template<typename R> int
glued_rep<R>::Type () {
  return GLUED_TYPE;
}

/*
template<typename R> void
glued_rep<R>::Print (ostream& out) {
  out << glued<R> (this);
}

template<typename R> ostream&
operator << (ostream& out, glued<R> x) {
  return out << x->data;
}
*/

/******************************************************************************
* Conversions
******************************************************************************/

/*
template<typename R> R view_as (generic g);

template <> int
view_as<int> (generic g) { 
  return (int)(mmx::as_int(g));
}

template <> char*
view_as<char*>(generic g) { 
  return as_charp(mmx::as_string(g));
}

template <> mmx::symbol
view_as<mmx::symbol>(generic g) { 
  return S(as_literal (g));
}

template <> mmx::alias
view_as<mmx::alias>(generic g) { 
  return as_alias(g);  
}
*/

template<typename R> inline glued<R>
as_glued (generic g) {
  if (gentype(g) != GLUED_TYPE)
    fatal_error ("generic is not an glued of required type",
		 "as_glued", "glued.hh");
  return glued<R> ((glued_rep<R>*) g.rep); 
}

template<typename R> R
view_as (generic g) {  
  return (as_glued<R>(g)->data);
}

static generic
as_glue_type (generic g) {
  fatal_error ("not yet implemented", "as_glue_type", "glued.hh");
  return generic ();
}

} //namespace mmx
#endif // __MMX_GLUED_HH
