
/******************************************************************************
* MODULE     : link.H
*              (C) B. Mourrain 
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/
#ifndef __MMX_EXTERN_HH
#define __MMX_EXTERN_HH
#include <generic.hh>
#include <single.hh>
#include <symbols.hh>
#include <literal.hh>
#include <alias.hh>
#include <function.hh>
#include <evaluator.hh>
#include <printer.hh>
#include <extern.hh>
#include <dlfcn.h>

#include <iostream>

namespace mmx {
template<class R> struct mmx_obj;
template<class R> mmx_obj<R> copy (mmx_obj<R> t);
template<class R> std::ostream& operator << (std::ostream&, mmx_obj<R>);
template<typename R>
struct mmx_rep: generic_rep {
private:
public:
  R data;  // the ostream

protected:
  int Type ();
  //  int Hash ();
  bool Equal (generic g) const;
  generic Copy ();
  int Arity ();
  void Print (std::ostream& out);

public:
  static const int TYPE=3000; // XXXX
  inline mmx_rep (const R & r): data (r) {}
  friend class mmx_obj<R>;
  friend mmx_obj<R> copy<> (mmx_obj<R> t);
  friend std::ostream& operator << <> (std::ostream& out, mmx_obj<R> a);
};

template<class R>
struct  mmx_obj {
protected:
  mmx_rep<R>* rep;
  inline mmx_obj<R> (mmx_rep<R>* r): rep (r) {}

public:
  inline mmx_obj<R> (const R & r): rep (new mmx_rep<R> (r)) {}
  friend mmx_obj<R> as_mmx_obj <> (generic g);

  inline operator R () { return rep->rep; }
  inline operator generic () { return generic (rep); }
  inline mmx_rep<R>* operator -> () { return rep; }
  inline mmx_rep<R>* operator -> () const { return rep; }

  inline bool operator == (const mmx_obj<R>& out) const
    { return (rep->data == out->data); }
  inline bool operator == (const mmx_obj<R>& out) 
    { return (rep->data == out->data); }
  inline bool operator != (const mmx_obj<R>& out) const
    { return (rep->data != out->data); }
  inline bool operator != (const mmx_obj<R>& out) 
    { return (rep->data != out->data); }
  friend mmx_obj<R> copy <> (mmx_obj<R> x);
  friend std::ostream& operator<< <> (std::ostream& out, mmx_obj<R> x);
  friend class mmx_rep<R>;
};

#include "link.C"

template<class R>  R view_as(generic g);

template <> int  view_as<int>(generic g) 
{ 
  return (int)(mmx::as_int(g));
}

template <> char* view_as<char*>(generic g) 
{ 
  return as_charp(mmx::as_string(g));
}

template <> mmx::symbol view_as<mmx::symbol>(generic g) 
{ 
  return S(as_literal (g));  
}

template <> mmx::alias view_as<mmx::alias>(generic g) 
{ 
  return as_alias(g);  
}

template<class R> R view_as(generic g) 
{  
  return (as_mmx_obj<R>(g)->data);
}
//----------------------------------------------------------------------
template<class T>
void set(mmx::alias v, const T & x) {
  v->ref = mmx::mmx_obj<T>(x);
}

template<>
void set(mmx::alias v, const int & x) {
  v->ref = mmx::single(x);
}

//----------------------------------------------------------------------
void new_function(char* s, mmx_function fn, generic type)
{
  std::cout<<s<<" : "<<cdr(car(cdr(type)))<<" -> "
    	   <<car(cdr(cdr(type)))
	   <<std::endl;
  mmeval->declare(symbol(s),value(function(s,fn),type));
}

void new_function(symbol s, mmx_function fn, generic type)
{
  std::cout<<s<<" : "<<cdr(car(cdr(type)))<<" -> "
    	   <<car(cdr(cdr(type)))
	   <<std::endl;
  mmeval->declare(s,value(function(s,fn),type));
}

void new_type(symbol x, generic v)
{
  std::cout<<x<<" : "<<mmx::ID_CLASS<<std::endl;
  mmeval->declare(x,value(v,mmx::ID_CLASS));
}
//----------------------------------------------------------------------
void  load_external(string name)
{
  string lib =  "mmx-"*name*".so";
  string rout= "init_" * name; // attention to mangling by C++
  char* _lib = as_charp (lib);
  char* _rout= as_charp (rout);
  void* handle= dlopen (_lib, RTLD_LAZY);//NOW | RTLD_GLOBAL);
  if (handle == NULL) {
    std::cerr<< dlerror ()<<std::endl;
    return;
  }
  else {
    void* f= dlsym (handle, _rout);
    if (f == NULL) {
      std::cerr<< "Symbol " * rout * " not defined in " * lib<<std::endl;
      return;
    }
    else {
      void (*init) (void)= (void (*) (void)) f;
      init ();
    }
  }
  dlclose(handle);  
}

} //namespace mmx
#endif // __MMX_EXTERN_HH
